#!/usr/bin/perl -w
use strict;

#
# constants
#
my $ENTITY_PREFIX = 'calc_';

#
# global variables
#
my %ENTITY;
my $BASE_ADDR = 0x08048000;

#
# parse_od
#
sub parse_od($)
{
  my $filename = shift;

  open(FILE, '<', $filename) || die "Can't open $filename: $!";
  LOOP: while(<FILE>)
  {
    chomp;
    my @number = split / /;
    my $start = hex($number[1]);

    ($start == 0x08048080) || die "$start != 0x08048080";
    $ENTITY{'evil_magic_start'} = '0x08048080';

    my $ofs = $start - $BASE_ADDR;
    ($ofs == 0x80) || die "$ofs != 0x80";
    $ENTITY{'evil_magic_ofs_D'} = $ofs;
    $ENTITY{'evil_magic_ofs_X'} = sprintf('0x%x', $ofs);

    last LOOP;
  }
  close FILE;
}

#
# set_num
#
sub set_num($$)
{
  my $base = shift;
  my $num = shift;

  $ENTITY{$base . '_D'} = $num;
  $ENTITY{$base . '_X'} = sprintf('0x%x', $num);
}

#
# parse_readelf_segment
#
sub parse_readelf_segment($$$$)
{
  my $filename = shift;
  my $base = shift;
  my $segment_name = shift;
  my $nr_load = shift;

  my @number = split / +/;
  my $virtaddr = hex($number[3]);
  my $filesiz = hex($number[5]);
  my $end = $virtaddr + $filesiz;

  # flags may contain spaces, e.g. "R E"
  # this will confuse the hell out of split(), so we use substr instead
  my $align = hex(substr($_, 68));
  if ($segment_name eq 'LOAD')
  {
    ($align == 4096) || die "$align != 4096 ($filename)";
  }

  # 'evil_magic_ofs_D' is set by parse_od()
  my $ofs = $ENTITY{$base . '_ofs_D'};
  if (defined($ofs))
  {
    my $codesize = $filesiz - $ofs;
    set_num($base . '_codesize', $codesize);

    # _size is set in parse_readelf()
    my $bloat = int($codesize) / int($ENTITY{$base . '_size'});
    $ENTITY{$base . '_bloat'} = sprintf("%.3f", $bloat);
    $ENTITY{$base . '_bloat_percent'} = sprintf("%.0f", $bloat * 100);
  }
  my $prev_base = $base . '_' . $segment_name . ($nr_load - 1);
  my $prev_end = $ENTITY{$prev_base . '_end_D'};
  $base .=  '_' . $segment_name . $nr_load;

  if (defined($prev_end))
  {
    my $diff = $virtaddr - $prev_end;
    set_num($base . '_diff', $diff);

    my $diff_align = $diff - $align;
    if ($diff_align < 0)
    {
      print "<!-- $diff_align &lt; 0 ($filename) -->\n";
    }
    else
    {
      set_num($base . '_diff_align', $diff_align);
    }
  }
  elsif ($nr_load == 1 && $segment_name eq 'LOAD')
  {
    ($virtaddr == $BASE_ADDR) || die "$virtaddr != $BASE_ADDR ($filename)";
  }
  set_num($base . '_offset', hex($number[2]));
  set_num($base . '_virtaddr', $virtaddr);
  set_num($base . '_filesiz', $filesiz);
  set_num($base . '_end', $end);
  set_num($base . '_align', $align);
}

#
# parse_readelf
#
sub parse_readelf
{
  my $filename = shift;
  my $base = shift;

  $filename =~ /\/([^\/]*)$/;
  $base = $1 if (!defined($base));
  my $nr_load = 1;

  my $entry_point;

  open(FILE, '<', $filename) || die "Can't open $filename: $!";
  while(<FILE>)
  {
    chomp;
    my @number = split / +/;
    if (m/^-rwx/)
    {
      $ENTITY{$base . '_size'} = $number[4];
    }
    elsif (m/^Entry point /)
    {
      $entry_point = hex($number[2]);
      set_num $base . '_entry_point', $entry_point;
      set_num $base . '_entry_point_start_ofs', $entry_point - $BASE_ADDR;
    }
    elsif (m/^\s*LOAD\s*/)
    {
      parse_readelf_segment $filename, $base, 'LOAD', $nr_load;
      if ($nr_load++ == 1)
      {
	set_num $base . '_entry_point_dist_to_end',
	  $ENTITY{$base . '_LOAD1_end_D'} - $entry_point;
      }
    }
    elsif (m/^\s*DYNAMIC\s*/)
    {
      parse_readelf_segment $filename, $base, 'DYNAMIC', 1;
    }
    elsif (m/^\s*NOTE\s*/)
    {
      parse_readelf_segment $filename, $base, 'NOTE', 1;
    }
  }
  close FILE;
}

#
# parse_ndisasm
#
sub parse_ndisasm($$)
{
  my $filename = shift;
  my $base = shift;
  my $nr_dword = 0;

  open(FILE, '<', $filename) || die "Can't open $filename: $!";
  LOOP: while(<FILE>)
  {
    my @number = split;
    if (m/\bpush dword 0x/)
    {
      set_num $base . '_ndisasm_dword' . $nr_dword++, hex($number[4]);
    }
    elsif (m/\bcall\b/)
    {
      set_num $base . '_ndisasm_call_addr', hex($number[3]);
      last LOOP;
    }
  }
  close FILE;
}

#
# parse_gdb
#
sub parse_gdb($$)
{
  my $filename = shift;
  my $base = shift;
  my $nr_dword = 0;
  my $nr_call = 0;

  open(FILE, '<', $filename) || die "Can't open $filename: $!";
  LOOP: while(<FILE>)
  {
    my @word = split;
    if (m/\bpush\s+0x/)
    {
      set_num $base . '_gdb_dword' . $nr_dword++, hex($word[3]);
    }
    elsif (m/\bcall\b/)
    {
      my $prefix = $base . '_gdb_call' . $nr_call++;
      set_num $prefix . '_addr', hex($word[3]);
      my $name = $word[4];
      $name =~ s/[<>]//g;
      $ENTITY{$prefix . '_name'} = $name;
    }
  }
  close FILE;
}

#
# parse_credits
#
sub parse_credits($)
{
  my $filename = shift;
  open(FILE, '<', $filename) || die "Can't open $filename: $!";
  my @credits = <FILE>;
  close FILE;

  my $line = join(', ', sort(@credits));
  $line =~ s/\s*\n//g;
  $ENTITY{'credits'} = $line;
}

#
# parse_ls
#
sub parse_ls($$)
{
  my $filename = shift;
  my $base = shift;

  open(FILE, '-|', 'ls -l ' . $filename) || die "Can't ls $filename: $!";
  LOOP: while(<FILE>)
  {
    chomp;
    my @number = split / +/;
    if (m/^-rw/)
    {
      $ENTITY{$base . '_size'} = $number[4];
    }
  }
  close FILE;
}

#
# main
#
$ENTITY{'base_addr'} = sprintf('0x%x', $BASE_ADDR);
parse_od 'out/evil_magic/od';
parse_readelf 'out/readelf/sh';
parse_readelf 'out/readelf/evil_magic';
parse_readelf 'out/entry_point/readelf', 'infone';
parse_ndisasm 'out/entry_point/ndisasm', 'sh';
parse_ndisasm 'out/entry_point/ndisasm_two', 'inftwo';
parse_ndisasm 'out/entry_point/ndisasm_three', 'infthree';
parse_gdb 'out/entry_point/gdb', 'sh';
parse_credits 'credits';
parse_ls 'tmp/one_step_closer/infection', 'infection';
parse_gdb 'out/doing_it_in_c/gdb_write_infection', 'write_infection';
parse_readelf 'out/additional_cs/readelf', 'additional_cs';

($ENTITY{'evil_magic_LOAD1_filesiz_D'} == 0x97)
|| die 'evil_magic_LOAD1_filesiz_D != 0x97';

set_num 'infone_size_delta',
  $ENTITY{'infone_size'} - $ENTITY{'sh_size'};
set_num 'infone_LOAD1_filesiz_delta',
  $ENTITY{'infone_LOAD1_filesiz_D'} - $ENTITY{'sh_LOAD1_filesiz_D'};
set_num 'infone_LOAD2_offset_delta',
  $ENTITY{'infone_LOAD2_offset_D'} - $ENTITY{'sh_LOAD2_offset_D'};
set_num 'infone_DYNAMIC1_offset_delta',
  $ENTITY{'infone_DYNAMIC1_offset_D'} - $ENTITY{'sh_DYNAMIC1_offset_D'};
set_num 'additional_cs_size_delta',
  $ENTITY{'additional_cs_size'} - $ENTITY{'sh_size'};

($ENTITY{'sh_gdb_call0_name'} eq '__libc_start_main')
|| die $ENTITY{'sh_gdb_call0_name'} . ' != __libc_start_main';

($ENTITY{'infection_size'} <= $ENTITY{'infone_NOTE1_filesiz_D'})
|| die $ENTITY{'infection_size'} . ' > ' . $ENTITY{'infone_NOTE1_filesiz_D'};

for my $name(keys(%ENTITY))
{
  printf "<!ENTITY %s%s \"%s\">\n", $ENTITY_PREFIX, $name, $ENTITY{$name};
}
